@@ -4,7 +4,7 @@ use warnings qw(FATAL all NONFATAL misc);
 use 5.010; no if $] >= 5.017011, warnings => "experimental";
 
 use Carp qw(carp croak confess);
-our $VERSION = '0.101_001';
+our $VERSION = '0.101_100';
 use mro 'c3';
 
 use Scalar::Util qw/weaken/;
@@ -14,6 +14,7 @@ use List::MoreUtils qw/uniq/;
 # YATT Internalへの Facade. YATT の初期化パラメータの保持者でもある。
 #
 use parent qw/YATT::Lite::Object File::Spec/;
+use YATT::Lite::Partial::MarkAfterNew -as_base;
 use YATT::Lite::MFields qw/YATT
 	      cf_dir
 	      cf_vfs cf_base
@@ -39,12 +40,17 @@ use YATT::Lite::MFields qw/YATT
 	      cf_lcmsg_sink
 	      cf_always_refresh_deps
 	      cf_no_mro_c3
+	      cf_render_as_bytes
 
 	      cf_default_lang
 
 	      cf_entns2vfs_item
 	      cf_import
               cf_match_argsroute_first
+              cf_stash_unknown_params_to
+              cf_body_argument
+              cf_body_argument_type
+	      cf_prefer_call_for_entity
 	    /;
 
 use constant DEBUG => $ENV{DEBUG_YATT_LITE};
@@ -54,7 +60,7 @@ MY->cf_mkaccessors(qw/app_name/);
 # Entities を多重継承する理由は import も継承したいから。
 # XXX: やっぱり、 YATT::Lite には固有の import を用意すべきではないか?
 #   yatt_default や cgen_perl を定義するための。
-use YATT::Lite::Entities -as_base, qw(*YATT *CON *SYS);
+use YATT::Lite::Entities -as_base, qw(*YATT *CON *SYS import);
 
 # For error, raise, DONE. This is inserted to ISA too.
 use YATT::Lite::Partial::ErrorReporter;
@@ -66,19 +72,25 @@ use YATT::Lite::Util qw/globref lexpand extname ckrequire terse_dump escape
 			look_for_globref
 			subname ckeval ckrequire
 			secure_text_plain
+                        raise_psgi_error
 			define_const
+                        add_entity_into
 		       /;
 
 sub Facade () {__PACKAGE__}
-sub default_app_ns {'MyApp'}
+sub default_app_ns {'MyYATT'}
 sub default_trans {'YATT::Lite::Core'}
 sub default_export {(shift->SUPER::default_export, qw(Entity *SYS *CON))}
 sub default_index_name { '' }
 sub default_ext_public {'yatt'}
 sub default_ext_private {'ytmpl'}
+sub default_body_argument { 'body' }
+sub default_body_argument_type { 'code' }
+sub default_output_encoding { 'utf-8' }
 
 sub with_system {
   (my MY $self, local $SYS, my $method) = splice @_, 0, 3;
+  local $YATT = $self;
   $self->$method(@_);
 }
 
@@ -88,6 +100,9 @@ sub after_new {
   $self->{cf_index_name} //= "";
   $self->{cf_ext_public} //= $self->default_ext_public;
   $self->{cf_ext_private} //= $self->default_ext_private;
+  $self->{cf_body_argument} //= $self->default_body_argument;
+  $self->{cf_body_argument_type} //= $self->default_body_argument_type;
+  $self->{cf_output_encoding} //= $self->default_output_encoding;
 }
 
 sub _after_after_new {
@@ -95,6 +110,18 @@ sub _after_after_new {
   weaken($self->{cf_factory});
 }
 
+sub rel_app_name {
+  (my MY $self) = @_;
+  unless ($SYS) {
+    Carp::croak "rel_app_name() is called without setting \$SYS!";
+  }
+  # XXX: This assumes $dir is somewhere under $app_root tree.
+  my $rel_app_name = substr($self->{cf_dir}, length($SYS->app_root));
+  # Make sure leading / is removed.
+  $rel_app_name =~ s|^/||;
+  $rel_app_name;
+}
+
 # XXX: kludge!
 sub find_neighbor_yatt {
   (my MY $self, my ($dir)) = @_;
@@ -175,20 +202,49 @@ sub handle {
 }
 
 sub render {
+  my MY $self = shift;
+  my $raw_bytes = $self->render_encoded(@_);
+  if ($self->{cf_render_as_bytes}) {
+    $raw_bytes
+  } else {
+    Encode::decode(utf8 => $raw_bytes);
+  }
+}
+
+sub render_encoded {
   my MY $self = shift;
   my $buffer; {
+    my $encName = $self->{cf_render_as_bytes} ? undef : $self->{cf_output_encoding};
     my $con = $SYS
-      ? $SYS->make_connection(undef, buffer => \$buffer, yatt => $self)
-	: ostream(\$buffer);
+      ? $SYS->make_connection(undef, buffer => \$buffer, yatt => $self,
+                              noheader => 1,
+                              encoding => $encName)
+	: ostream(\$buffer, $encName ? ":encoding($encName)" : "");
     $self->render_into($con, @_);
   }
   $buffer;
 }
 
 sub render_into {
-  local ($YATT, $CON) = splice @_, 0, 2;
-  $YATT->open_trans->render_into($CON, @_);
-  try_invoke($CON, 'flush_headers');
+  (my MY $self, my ($con, $file, $args)) = @_;
+  local ($YATT, $CON) = ($self, $con);
+  if (ref $file eq 'ARRAY') {
+    # [$file, $type, $item]
+    $self->open_trans->render_into($CON, $file, $args);
+  } else {
+    $self->raw_render_into(
+      $con,
+      YATT::Lite::Util::rootname($file || $self->{cf_index_name}),
+      $args,
+    );
+  }
+  try_invoke($con, 'flush_headers');
+}
+
+sub raw_render_into {
+  (local ($YATT), my ($con, $file, $args)) = @_;
+  my ($part, $sub, $pkg) = $YATT->open_trans->find_part_renderer($file);
+  $sub->($pkg, $con, $part->reorder_hash_params($args));
 }
 
 sub find_handler {
@@ -216,8 +272,8 @@ sub _handle_yatt {
 
 sub _handle_ytmpl {
   (my MY $self, my ($con, $file)) = @_;
-  # XXX: http result code:
-  print $con "Forbidden filetype: $file";
+
+  $self->raise_psgi_error(403, "Forbidden filetype: $file");
 }
 
 #----------------------------------------
@@ -227,7 +283,7 @@ sub prepare_part_handler {
 
   my $trans = $self->open_trans;
 
-  my $mapped = [$file, my ($type, $item) = $self->parse_request_sigil($con)];
+  my $mapped = [$file, my ($type, $item) = $con->sigil_type_item()];
   if (not $self->{cf_dont_debug_param}
       and -e ".htdebug_param") {
     $self->dump($mapped, [map {[$_ => $con->param($_)]} $con->param]);
@@ -240,54 +296,16 @@ sub prepare_part_handler {
     croak $self->error(q|Forbidden request %s|, terse_dump($mapped));
   }
 
-  my @args; @args = $part->reorder_cgi_params($con)
+  my @args; @args = $trans->reorder_cgi_params($part, $con)
     unless $self->{cf_dont_map_args} || $part->isa($trans->Action);
 
   ($part, $sub, $pkg, \@args);
 }
 
+# XXX: deprecated
 sub parse_request_sigil {
   (my MY $self, my ($con)) = @_;
-  my ($subpage, $action);
-  # XXX: url_param
-  foreach my $name (grep {defined} $con->param()) {
-    my ($sigil, $word) = $name =~ /^([~!])(\1|\w*)$/
-      or next;
-    # If $name in ('~~', '!!'), use value.
-    my $new = $word eq $sigil ? $con->param($name) : $word;
-    # else use $word from ~$word.
-    # Note: $word may eq ''. This is for render_/action_.
-    given ($sigil) {
-      when ('~') {
-	if (defined $subpage) {
-	  $self->error("Duplicate subpage request! %s vs %s"
-		       , $subpage, $new);
-	}
-	$subpage = $new;
-      }
-      when ('!') {
-	if (defined $action) {
-	  $self->error("Duplicate action! %s vs %s"
-		       , $action, $new);
-	}
-	$action = $new;
-      }
-      default {
-	croak "Really?";
-      }
-    }
-  }
-  if (defined $subpage and defined $action) {
-    # XXX: Reserved for future use.
-    $self->error("Can't use subpage and action at one time: %s vs %s"
-		 , $subpage, $action);
-  } elsif (defined $subpage) {
-    (page => $subpage);
-  } elsif (defined $action) {
-    (action => $action);
-  } else {
-    ();
-  }
+  $con->sigil_type_item();
 }
 
 sub cut_ext {
@@ -373,6 +391,10 @@ sub _cf_delegates {
      no_mro_c3
      import
      match_argsroute_first
+     stash_unknown_params_to
+     body_argument
+     body_argument_type
+     prefer_call_for_entity
     )
 }
 
@@ -557,10 +579,9 @@ sub define_Entity {
   unless (*{$ent}{CODE}) {
     *$ent = sub {
       my ($name, $sub) = @_;
-      my $longname = join "::", $destns, "entity_$name";
-      subname($longname, $sub);
-      print STDERR "defining entity_$name in $destns\n" if DEBUG;
-      *{globref($destns, "entity_$name")} = $sub;
+      if ($myPack->add_entity_into($destns, $name, $sub)) {
+        print STDERR "# defined entity $name in $destns\n" if DEBUG;
+      }
     };
   }
 
@@ -682,6 +703,11 @@ sub default_lang {
 
 #========================================
 # Delegation to the core(Translator, which is useless for non-templating.)
+#
+# Note: these wrappers do not call YATT::Lite::VFS->reset_refresh_mark.
+# This means once templates is compiled, it is not updated even if
+# they are modified. You can avoid this behavior by
+# directly using $YATT->open_trans->xxx instead.
 #========================================
 foreach
   (qw/find_part
@@ -693,6 +719,8 @@ foreach
       ensure_parsed
 
       list_items
+      list_base
+      list_internal_base_folders
 
       add_to
     /
@@ -753,7 +781,13 @@ sub YATT::Lite::EntNS::entity_mkhidden {
         _hidden_input(escape($name), $v);
       }
     } $CON->multi_param($name);
-  } @_;
+  } do {
+    if (@_) {
+      @_;
+    } else {
+      $CON->param
+    }
+  };
 };
 
 sub _hidden_input {
@@ -768,6 +802,11 @@ sub YATT::Lite::EntNS::entity_file_rootname {
   $fn;
 };
 
+sub YATT::Lite::EntNS::entity_app_name {
+  my ($this) = @_;
+  $YATT->cget('app_name');
+};
+
 #----------------------------------------
 use YATT::Lite::Breakpoint ();
 YATT::Lite::Breakpoint::break_load_facade();
